// Interpolator.cpp : Defines the entry point for the console application.
//

#include "stdafx.h"
#include <stdlib.h>
#include <malloc.h>
#include <conio.h>

#define FREE(x) (free(x), x=NULL)

double abs(double val)
{
	if ( val < 0 ) return 0-val; else return val;
}

class Vector {
private:
	double * v;
	int size;
	bool horiz;

public:
	~Vector() { FREE(v); }
	Vector() { v=NULL; }
	Vector(int s, bool h) { v=(double*)malloc(sizeof(double)*s); size=s; horiz=h; }
	void init(int s, bool h) { if ( v ) FREE(v); v=(double*)malloc(sizeof(double)*s); size=s; horiz=h; }
	
	void disp() { for ( int p=0; p<size; p++ ) { printf("% 6.5f ", v[p]); if ( !horiz ) printf("\n"); } }
	double & operator[] (int p) { return v[p-1]; }
	void invert() { double temp; for ( int t=(size>>1)-1; t>=0; t-- ) { temp=v[t]; v[t]=v[size-t-1]; v[size-t-1]=temp; } }

/*	double set(int p, double val) { v[p-1]=val; return v[p-1]; }
	void transpose() { if ( horiz ) horiz=false; else horiz=true; }*/
};

class Matrix {
private:
	double * im;
	int imheight;
	int imwidth;

public:
	~Matrix() { if ( im ) FREE(im); }
	Matrix() { im=NULL; }
	Matrix(int h, int w) { im=(double*)malloc(sizeof(double)*w*h); imheight=h; imwidth=w; }
	void init(int h, int w) { if ( im ) FREE(im); im=(double*)malloc(sizeof(double)*w*h); imheight=h; imwidth=w; }
	
	void disp();
	int height() { return imheight; }
	int width() { return imwidth; }
	double e(int y, int x) { return *(im+((y-1)*imwidth)+(x-1)); }

	double set(int y, int x, double val) { *(im+((y-1)*imwidth)+(x-1))=val; return val; }

//	void transpose();
	Matrix & operator = (Matrix &m2);
	Matrix & operator = (double val);
/*	Matrix & operator += (Matrix &m2); 
	Matrix & operator -= (Matrix &m2);
	Matrix & operator *= (Matrix &m2);*/
};

void Matrix::disp()
{
	int y, x;

	for(y=0; y<imheight; y++) {
		for(x=0; x<imwidth; x++) {
			printf("% 6.5f ", *(im+(y*imwidth)+x));
		}
		printf("\n");
	}
}
/*
void Matrix::transpose()
{
	double * m2=(double*)malloc(sizeof(double)*imheight*imwidth);
	int t;

	int y, x;

	for(y=0; y<imheight; y++) {
		for(x=0; x<imwidth; x++) {
			*(m2+(x*imheight)+y)=*(im+(y*imwidth)+x);
		}
	}
	t=imwidth;
	imwidth=imheight;
	imheight=t;
	FREE(im);
	im=m2;
}
*/
Matrix & Matrix::operator = (Matrix &m2)
{
	if ( im ) FREE(im);

	imheight=m2.height();
	imwidth=m2.width();

	im=(double*)malloc(sizeof(double)*imheight*imwidth);

	int y, x;

	for(y=0; y<imheight; y++) {
		for(x=0; x<imwidth; x++) {
			*(im+(y*imwidth)+x)=m2.e(y+1, x+1);
		}
	}

	return *this;
}

Matrix & Matrix::operator = (double val)
{
	int y, x;

	for(y=0; y<imheight; y++) {
		for(x=0; x<imwidth; x++) {
			*(im+(y*imwidth)+x)=val;
		}
	}

	return *this;
}
/*
Matrix & Matrix::operator += (Matrix &m2)
{
	int h=(imheight<m2.height()) ? imheight : m2.height();
	int w=(imwidth<m2.width()) ? imwidth : m2.width();

	int y, x;

	for(y=0; y<h; y++) {
		for(x=0; x<w; x++) {
			*(im+(y*imwidth)+x) += m2.e(y+1, x+1);
		}
	}

	return *this;
}

Matrix & Matrix::operator -= (Matrix &m2)
{
	int h=(imheight<m2.height()) ? imheight : m2.height();
	int w=(imwidth<m2.width()) ? imwidth : m2.width();

	int y, x;

	for(y=0; y<h; y++) {
		for(x=0; x<w; x++) {
			*(im+(y*imwidth)+x) -= m2.e(y+1, x+1);
		}
	}

	return *this;
}

Matrix & Matrix::operator *= (Matrix &m2)
{
	if ( imwidth != m2.height() ) {
		printf("not m*n, n*p matrices!");
		return *this;
	}

	int h=imheight;
	int w=m2.width();
	int n=imwidth;

	double *m=(double*)calloc(h*w, sizeof(double));

	int y, x, t;

	for(y=0; y<h; y++) {
		for(x=0; x<w; x++) {
			for(t=0; t<n; t++) {
				*(m+(y*w)+x) += *(im+(y*imwidth)+t) * m2.e(t+1, x+1);
			}
		}
	}

	imheight=h;
	imwidth=w;

	FREE(im);
	im=m;

	return *this;
}
*/
void LUP_Decomposition(Matrix &A, Matrix &L, Matrix &U, Vector &pi)
{
	int n;
	int i, k, k2, j;
	double t, p;

	Matrix A2;
	A2=A;

	n=A2.height();

	pi.init(n, true);

	for(i=1; i<=n; i++)
		pi[i]=i;

	for(k=1; k<n; k++) {
		p=0;
		for(i=k; i<=n; i++) {
			if ( abs(A2.e(i,k)) > p ) {
				p=abs(A2.e(i,k));
				k2=i;
			}
		}
		if ( p == 0 ) {
			printf("Singular matrix");
			exit(0);
		}
		t=pi[k];
		pi[k]=pi[k2];
		pi[k2]=t;
		for(i=1; i<=n; i++) {
			t=A2.e(k,i);
			A2.set(k,i, A2.e(k2,i));
			A2.set(k2,i, t);
		}
		for(i=k+1; i<=n; i++) {
			A2.set(i,k, A2.e(i,k)/A2.e(k,k));
			for(j=k+1; j<=n; j++)
				A2.set(i,j, A2.e(i,j) - (A2.e(i,k)*A2.e(k,j)));
		}
	}

	L.init(n,n);
	U.init(n,n);
	for(i=1; i<=n; i++) {
		for(j=1; j<i; j++) {
			L.set(i,j, A2.e(i,j));
			U.set(i,j, 0);
		}
		for(j=i; j<=n; j++) {
			L.set(i,j, 0);
			U.set(i,j, A2.e(i,j));
		}
		L.set(i,i, 1);
	}
}

void LUP_Solve(Matrix &L, Matrix &U, Vector &pi, Vector &b, Vector &x)
{
	int n, i, j;
	Vector y;
	
	n=L.height();
	y.init(n, true);
	x.init(n, false);

	for(i=1; i<=n; i++) {
		y[i]=b[(int)pi[i]];
		for(j=1; j<i; j++)
			y[i]-=L.e(i,j)*y[j];
	}
	for(i=n; i>=1; i--) {
		x[i]=y[i];
		for(j=i+1; j<=n; j++)
			x[i]-=U.e(i,j)*x[j];
		x[i]/=U.e(i,i);
	}
}

int main(int argc, char* argv[])
{
	Matrix A(3, 3);
	Matrix L;
	Matrix U;
	Vector pi, b(3, false), x;

	A.set(1,1, 1);
	A.set(1,2, 2);
	A.set(1,3, 0);
	A.set(2,1, 3);
	A.set(2,2, 5);
	A.set(2,3, 4);
	A.set(3,1, 5);
	A.set(3,2, 6);
	A.set(3,3, 3);

	b[1]=0.1;
	b[2]=12.5;
	b[3]=10.3;

	printf("\nBeginning\n");
	printf("=========\n\n");

	printf("A:\n");
	A.disp();
	printf("\n");

	printf("b:\n");
	b.disp();
	printf("\n");

	_getch();

	printf("\nLUP Decomposition\n");
	printf("=================\n\n");

	LUP_Decomposition(A, L, U, pi);

	printf("A:\n");
	A.disp();
	printf("\n");

	printf("L:\n");
	L.disp();
	printf("\n");

	printf("U:\n");
	U.disp();
	printf("\n");

	printf("pi:\n");
	pi.disp();
	printf("\n");

	_getch();

	printf("\nLUP Solve\n");
	printf("=========\n\n");

	LUP_Solve(L, U, pi, b, x);

	printf("x:\n");
	x.disp();
	printf("\n");

	return 0;
}

